# ******************* DEBUG ******************** #
#' output to screen
#' @param ... fed in to \code{cat}
#' @seealso verbosePrintf
#' @family debug
#' @export
verbosePrint <- function(...) cat(...)

#' output to screen, like sprintf.
#' @param ... fed in to \code{sprintf} which is then \code{cat}
#' @seealso verbosePrint
#' @seealso \code{\link[base]{sprintf}}
#' @family debug
#' @export
verbosePrintf <- function(...) cat(sprintf(...))

#' throw a formatted warning.
#' Essentially calls \code{warning(sprintf(...))}
#' @param ... arguments to sprintf
#' @param call. logical, indicating if call should become part of the
#' warning message (see \code{\link[base]{warning}})
#' @param immediate. logical, indicating of the call should be output
#' immediately, even if \code{getOption("warn")<=0} 
#' @param domain see \code{\link[base]{warning}}
#' @note See \code{\link[base]{warning}} for explanation of parameters.
#' @family debug
#' @seealso \code{\link[base]{warning}}, \code{\link[base]{sprintf}}
#' @export
warningf <- function(...,call.=TRUE, immediate.=FALSE,domain=NULL) 
    warning(sprintf(...),call.=call.,immediate.=immediate.,domain=domain)

#' throw a formated error
#' Essentially calls \code{stop(sprintf(...))}
#' @param ... arguments to sprintf
#' @param call. logical, indicating if call should become part of the
#' error message (see \code{\link[base]{stop}})
#' @param domain see \code{\link[base]{stop}}
#' See \code{\link[base]{stop}} for the explanation of the other parameters.
#' @family debug
#' @seealso \code{\link[base]{stop}}, \code{\link[base]{sprintf}}
#' @export
stopf <- function(...,call.=TRUE,domain=NULL)
    stop(sprintf(...),call.=call.,domain=domain)


#' Catch *and* save both errors and warnings, and in the case of
#' a warning, also keep the computed result.
#'
#' @title tryCatch both warnings and errors
#' @param expr the expression to evaluate
#' @return a list with 'value' and 'warning', where 
#'  'value' may be an error caught.
#' @author Martin Maechler
#' @family debug
#' @export
tryCatch.W.E <- function(expr) 
{
    W <- NULL 
    w.handler <- function(w){ # warning handler

    W <<- w
    invokeRestart("muffleWarning")

    } 
    list(value = withCallingHandlers(tryCatch(expr, error = function(e) e),
                     warning = w.handler),
     warning = W)

}

# ********************* DEMOS ********************** #
#' convert a delimited string to a vector.
#' @param x the delimited string.
#' @details Essentially calls c(...) on the arguments, after
#'         delimiting them.
#'         Delimited by 'x' or ','
#' @examples 
#' evaluateSplit('1,2,3')       # returns c(1,2,3)
#' evaluateSplit('640x480') # returns c(640,480)
#' @family demo
# NOTE:
# gWidgetstcltk has this problem where if you do
# svalue(obj) AND if you created obj with a coerce.with, it will 
#  call coerce.with TWICE:
# it does  coerce.with(coerce.with(string_value_of(obj))).
# Hence the reason for the curious paste(x,collapse=',').
#' @export
evaluateSplit <- function(x) 
        evaluate(sprintf('c(%s)',gsub('x',',',paste(x,collapse=','))))

#' evaluate string as an R expression
#' @param x string with command to be evaluated
#' @return result of evaluating x as if it were an R expression
#' @family demo
#' @export
evaluate <- function(x) eval(parse(text=x),parent.frame())

#' *P*rints a string and then *Ev*aluates it
#'
#' Useful for interactive demos.
#' @param str the command to evaluate
#' @return the result of the expression
#' @family demo
#' @export
PEv <- function( str ) 
{ 
    cat(str,"\n\n"); 
    evaluate(str)
}

#' Pauses evaluation, waiting for user input.
#'
#' This displays "press ENTER to continue" and waits until the user has done so.
#' Useful for interactive demos.
#' @family demo
#' @export
wait <- function() 
{ 
    cat("                                                              press ENTER to continue\n");
    scan(file="", quiet=T, what = character(0), nlines=1)
}

#' *D*isplays a command, *W*aits, and *E*xecutes.
#'
#' This displays a command, waits for the user to press enter, 
#'  and then executes the command.
#' Useful for interactive demos.
#' @param str the command to evaluate
#' @return the result of the expression
#' @family demo
#' @export
DWE <- function( str )
{
    cat( str,"\n\n" )
    cat("                                                              press ENTER to evaluate\n");
    scan(file="", quiet=T, what = character(0), nlines=1)
    eval(parse(text=str),parent.frame())    
}


