# *************************** Files **************************** #
#' retrieves path to system executable.
#'
#' This essentially emulates the unix "which".
#' @param PROG_NAMES a vector of program names to search for.
#' @return a vector of absolute paths to the executables specified.
#' @export
getProgram <- function(PROG_NAMES)
{
    NAMES <- unname(Sys.which(PROG_NAMES))
    NAME <- NAMES[NAMES!='']
    if ( length(NAME)>0 )
        return(NAME)
    else
        return(NULL)
}

#' Text-based interactive file selection.
#' @param root the root directory to explore
#'             (default current working directory)
#' @param multiple boolean specifying whether to allow 
#'                 multiple files to be selected
#' @param title what gets printed before the menu.
#' @return character vector of selected files.
#' @examples 
#' \dontrun{
#' fileList <- file.browse()
#' }
#' @family file
#' @seealso \code{\link{dir.browse}} for a similar functionality for
#' directories.
#' @export
file.browse <- function (root=getwd(), multiple=F,
               title=sprintf('Select file(s) (0 to quit) in folder %s',
                               root)) 
{
    # .. and list.files(root)
    x <- c( dirname(normalizePath(root)), list.files(root,full.names=T) )
    isdir <- file.info(x)$isdir
    obj <- sort(isdir,index.return=T,decreasing=T)
    isdir <- obj$x
    x <- x[obj$ix]
    lbls <- sprintf('%s%s',basename(x),ifelse(isdir,'/',''))
    lbls[1] <- sprintf('../ (%s)', basename(x[1]))

    files <- c()
    sel = -1
    while ( TRUE ) {
        sel <- menu(lbls,title=title)
        if (sel == 0 ) 
            break
        if (isdir[sel]) {
            # directory, browse further
            files <- c(files, file.browse( x[sel], multiple ))
            break
        } else {
            # file, add to list
            files <- c(files,x[sel])
            if ( !multiple ) 
                break
            # remove selected file from choices
            lbls <- lbls[-sel]
            x <- x[-sel]
            isdir <- isdir[-sel]
        }
    }
    return(files)
}

#' Text-based interactive directory selection.
#' @param root the root directory to explore
#'             (default current working directory)
#' @param multiple boolean specifying whether to allow 
#'                 multiple directories to be selected
#' @param title what gets printed before the menu.
#' @param .select *internal* argument, how the function works out whether
#'                you're trying to select a directory or browse it.
#' @return character vector of selected directories.
#' @note This could be a bit buggy, due to the fact that it's a little 
#' hard to tell whether a user wishes to browse into a directory, or
#' select it.
#' @examples 
#' \dontrun{
#' d <- dir.browse()
#' }
#' @family file
#' @seealso \code{\link{file.browse}} for a similar functionality for
#' directories.
#' @export
dir.browse <- function( root=getwd(), multiple=F, 
                       title=sprintf('Select a directory'), .select=FALSE)
{
    # .. and list.files(root)
    x <- c( dirname(normalizePath(root)), list.dirs(root,full.names=T) )
    lbls <- sprintf('%s/',basename(x))
    lbls[1] <- sprintf('../ (%s)', basename(x[1]))
    cd <- which(x=='.')
    if ( length(cd)>0 ) lbls[cd] <- sprintf('./ (%s)', basename(x[cd])) 
    n <- length(lbls)+1
    lbls[n] <- ifelse(.select,
                       'Go back to browsing mode',
                       'Add a directory')
    dirs <- c()
    sel = -1
    while ( TRUE ) {
        sel <- menu(lbls,title=title)
        if (sel == 0 ) 
            break
        # switch between browsing and adding
        if ( sel == n ) {
            .select <- !.select
            next
        }

        # otherwise add to list OR browse to that dir
        if ( .select && sel != 1 ) {
            # add to list, unless they chose .. because that's too hard!
            dirs <- c(dirs,x[sel])
            if ( !multiple )
                break

            # remove selected dir from choices
            # Urgh no, this will end up 
            # lbls <- lbls[-sel]
            # x <- x[-sel]
        } else {
            # they're browsing, navigate there.
            dirs <- c(dirs,
                      dir.browse(x[sel],multiple,title,.select=FALSE))
            break
        }
    }
    return(unique(dirs))
}


if (!exists('list.dirs',mode='function')){
#' lists all directories
#' @param ... See \code{\link[base]{list.files}}
#' @return vector of directories
#' @seealso \code{\link[base]{list.files}}
#' @family file
#' @export
list.dirs <- function(...) {
    tmp <- list.files(...)
    return(tmp[file.info(tmp)$isdir])
}
}

#' get the extension of a path
#' @param path the path to get the extension of
#' @param preserveCase boolean, default FALSE. 
#'                     Whether or not to preserve the case of the extension.
#'                     If FALSE, returns lowercase.
#' @return the extension of the path, being the text from the last '.' to the end.
#'         If no extension, returns ''.
#' @family file
#' @export
getExtension <- function( path, preserveCase=FALSE ) 
{
    parts <- strsplit(basename(path), ".", fixed=TRUE)[[1]]
    if ( length(parts)==1 && basename(path)==parts[1] ) return("")
    if ( preserveCase ) {
        return( parts[length(parts)] )
    } else {
        return( tolower(parts[length(parts)]) )
    }
}

#' get the file name of a path not including the extension.
#' @param fPath the path to get the file name of
#' @return The part of the file path from the last occuring '/', not including the extension.
#' @family file
#' @export
getFilename <- function( fPath )
# returns the file name not including the extension.
{
    tmp        <- basename(fPath)                           # includes extension
    parts      <- strsplit(tmp, ".", fixed=TRUE)            # split the fname from the ext

    # if there are multiple '.' in file name assume just the last is the extension.
    return(vapply(parts,function(x) if ( length(x)==1 ) return(x) else return(paste(x[-length(x)],collapse=".")),"returnTemplate"))
}

# TODO: overwrite=T doesn't belong here, it belongs in a writing function@
# TODO: createFilename has overwrite=F

#' create a file path based off a given format.
#'
#' @param format    A file format optionally containing the following
#'                  tokens, which will be replaced from the "base" and 
#'                  "fPath" arguments.
#'  \describe{
#'     item{\%d}{the parent directory of the file in fPath}
#'     item{\%D}{dirname(fPath)}
#'     item{\%e}{the extension of the file in fPath (default "tif")}
#'     item{\%f}{the file name of the file in fPath}
#'     item{\%B}{the "base" parameter supplied in the arguments.}
#'   }
#' @param fPath     A filepath to grab "\%d", "\%D", etc from.
#' @param base      what to be used as "\%B" in "format".
#' @param overwrite boolean, default FALSE, whether to return a filename
#'                  that doesn't exist in the file system by appending
#'                  numbers to the end (filename_1, filename_2, ...)
#' @return a file name with all the tokens substituted in.
#' @examples
#' createFilename('\%B/\%d/\%f_new.\%e',fPath='/path/to/file.txt',
#'                 base='/new/root/dir')
#'  # returns '/new/root/dir/to/file_new.txt'
#' createFilename('\%D/\%f_smaller.jpg',fPath='~/pictures/img.jpg')
#'  # returns '~/pictures/img_smaller.jpg'
#' @family file
#' @export
createFilename <- function(format,fPath=format,base="",overwrite=FALSE) 
{
    # %d will get the directory directly above the file name
    parts <- strsplit(fPath,"/")[[1]] 
    dName <- ifelse( length(parts)==1,"",parts[length(parts)-1] )
    dirParts <- parts[1:(length(parts)-1)]
    parts <- strsplit( parts[length(parts)], ".", fixed=TRUE)[[1]] # split the fname from the ext
    ext   <- ifelse( length(parts)==1, "tif", parts[length(parts)] )
    fName <- ifelse( length(parts)==1, parts[1], paste(parts[-length(parts)],collapse=".") )

    outName  <- gsub( "%B", base, format )
    outName  <- gsub( "%d", dName, outName )
    outName  <- gsub( "%f", fName, outName )
    outName  <- gsub( "%e", ext,   outName )
    outName  <- gsub( "%D", dirname(fPath), outName )
  
     
    # NOTE: %d is the parent directory, %nd is [n] the parent directories. 
    # (so ../data/123/456/img.tif has %d 456, and %2d is 123/456, etc. 
    # See if there's any of %[0-9]d
    if ( grepl("(?<=%)[0-9](?=d)",outName,perl=T) ) {
        # find what the number is. grep's too dumb to return things in ()
        # and I don't want to load other libraries, so I'll be a little silly
        # and strip everything that's NOT %[0-9]d. 
        # The following RELIES on there being no more '%\d' left besides '%\dd'.
        num <- as.integer(gsub("[^0-9]|(?<!%)[0-9](?!=d)","",outName,perl=T))        
        n    <- length(dirParts)
        dirs <- ifelse( num > n,
                        paste(dirParts[1:n],collapse="/"),
                        paste(dirParts[(n-num+1):n],collapse="/")
                      )
        outName <- gsub( "%[0-9]d", dirs, outName )
    }
    
    outName  <- gsub( "//", "/",   outName )

    # check for existence. Try to append a number to the end if it exists already.
    # alternatively, use tempfile(pattern=getFilename(outName),tmpdir=dirname(outName),fileext=getFileExtension(outName))
    #  which will add a random string.
    # ** THE SEQUENTIAL NUMBER SUFFIX MODE IS EXPERIMENTAL AND NOT GUARANTEED TO WORK
    outName <- getUniqueFilename(outName,overwrite=overwrite)
    return( outName )
}

#' Generates a unique file name by incrementing a counter.
#' 
#' @param fPath     file path to make unique
#' @param overwrite boolean, default FALSE, whether to generate a file name
#'                  that is the name of an already-existing file, i.e.
#'                  *don't* make the filename unique.
#' @return unique file name, consisting of the original with a number
#'          appended (filename1.txt, filename2.txt, etc).
#' @examples
#' getUniqueFilename('foo.txt')
#'  # if foo.txt exists, returns 'foo1.txt'
#'  # if foo1.txt exists, returns 'foo2.txt'. etc.
#' @family file
#' @export
getUniqueFilename <- function(fPath,overwrite=F) 
{
    if ( overwrite ) return(fPath)

    if ( !is.na( file.info(fPath)$size ) ) { # already exists.
        fName <- getFilename(fPath)
        ext   <- getExtension(fPath)
        if ( ext != "" ) ext <- sprintf('.%s',ext)

        fList <- list.files( dirname(fPath), sprintf("^%s.*%s",fName,ext) ) 
        if ( length(fList) > 0 ) {
            newFName <- make.unique( c(getFilename(fList),fName), sep="" )[length(fList)+1]
            outName <- sprintf('%s/%s%s',dirname(fPath),newFName,ext)
            # sfx <- max(as.integer(gsub("[^0-9]","",fList)))+1 # call it experiment5.Rdata if 1--4 exist, etc.
            cat(sprintf('  warning: createFilename: file %s exists already, saving as %s\n',fPath,outName))
        }
        fPath <- outName
        #outName <- createFilename(sprintf("%%D/%%f%i.%%e",as.integer(sfx)),outName)
    } 
    return( fPath )
}

#' makes a directory, if it doesn't already exist.
#' @param dirpaths vector of paths to create (or a path to create)
#' @family file
#' @export
makeIfNotExists <- function( dirpaths ) 
# Will make directory dirpath if it doesn't exist
{
    for ( dirpath in dirpaths ) {
        if ( is.na(file.info(dirpath)$size) ) {
            cat("directory",dirpath,"does not exist, creating.\n")
            dir.create(dirpath,recursive=TRUE)
        }
    }
}

